@@ -1,21 +1,19 @@
-from sklearn.base import BaseEstimator
-from sklearn.exceptions import NotFittedError
+from pandas import CategoricalDtype
 try:
-	from sklearn.compose import ColumnTransformer
+	from sklearn_pandas import DataFrameMapper
 except ImportError:
-	class ColumnTransformer:
+	class DataFrameMapper(object):
 		pass
-try:
-	from sklearn.feature_selection._base import SelectorMixin
-except ImportError:
-	from sklearn.feature_selection.base import SelectorMixin
-from sklearn.feature_selection import SelectFromModel
+from sklearn.base import BaseEstimator
+from sklearn.compose import ColumnTransformer
+from sklearn.exceptions import NotFittedError
+from sklearn.feature_selection import SelectFromModel, SelectorMixin
 from sklearn.pipeline import FeatureUnion, Pipeline
-from sklearn_pandas import DataFrameMapper
 from sklearn2pmml.resources import _package_classpath
 from subprocess import PIPE, Popen
 from zipfile import ZipFile
 
+import dill
 import joblib
 import numpy
 import os
@@ -23,8 +21,8 @@ import pandas
 import platform
 import re
 import sklearn
-import sklearn_pandas
 import tempfile
+import warnings
 
 from .metadata import __copyright__, __license__, __version__
 from .pipeline import PMMLPipeline
@@ -41,6 +39,23 @@ def _is_pandas_categorical(dtype):
 		return dtype.name == "category"
 	return False
 
+def _is_proto_pandas_categorical(dtype):
+	if isinstance(dtype, str) and dtype == "category":
+		return True
+	if isinstance(dtype, CategoricalDtype):
+		return dtype.categories is None
+	return False
+
+def _is_ordinal(dtype):
+	if _is_pandas_ordinal(dtype):
+		return True
+	return False
+
+def _is_pandas_ordinal(dtype):
+	if isinstance(dtype, CategoricalDtype):
+		return dtype.ordered
+	return False
+
 class EstimatorProxy(BaseEstimator):
 
 	def __init__(self, estimator, attr_names = ["feature_importances_"]):
@@ -56,6 +71,10 @@ class EstimatorProxy(BaseEstimator):
 			if hasattr(self.estimator, attr_name):
 				setattr(self, attr_name, getattr(self.estimator, attr_name))
 
+	@property
+	def classes_(self):
+		return self.estimator.classes_
+
 	def fit(self, X, y = None, **fit_params):
 		self.estimator.fit(X, y, **fit_params)
 		self._copy_attrs()
@@ -90,45 +109,37 @@ class SelectorProxy(BaseEstimator):
 	def transform(self, X):
 		return self.selector.transform(X)
 
-def _get_steps(obj):
-	if isinstance(obj, Pipeline):
-		return obj.steps
-	elif isinstance(obj, BaseEstimator):
-		return [("estimator", obj)]
-	else:
-		raise TypeError("The object is not an instance of {0}".format(BaseEstimator.__name__))
-
-def _filter(obj):
+def _escape(obj, escape_func):
 	if isinstance(obj, DataFrameMapper):
-		obj.features = _filter_steps(obj.features)
+		obj.features = _escape_steps(obj.features, escape_func = escape_func)
 		if hasattr(obj, "built_features"):
 			if obj.built_features is not None:
-				obj.built_features = _filter_steps(obj.built_features)
+				obj.built_features = _escape_steps(obj.built_features, escape_func = escape_func)
 	elif isinstance(obj, ColumnTransformer):
-		obj.transformers = _filter_steps(obj.transformers)
-		obj.remainder = _filter(obj.remainder)
+		obj.transformers = _escape_steps(obj.transformers, escape_func = escape_func)
+		obj.remainder = escape_func(obj.remainder, escape_func = escape_func)
 		if hasattr(obj, "transformers_"):
-			obj.transformers_ = _filter_steps(obj.transformers_)
+			obj.transformers_ = _escape_steps(obj.transformers_, escape_func = escape_func)
 	elif isinstance(obj, FeatureUnion):
-		obj.transformer_list = _filter_steps(obj.transformer_list)
+		obj.transformer_list = _escape_steps(obj.transformer_list, escape_func = escape_func)
 	elif isinstance(obj, Pipeline):
-		obj.steps = _filter_steps(obj.steps)
+		obj.steps = _escape_steps(obj.steps, escape_func = escape_func)
 	elif isinstance(obj, SelectorMixin):
 		return SelectorProxy(obj)
 	elif isinstance(obj, list):
-		return [_filter(e) for e in obj]
+		return [escape_func(e, escape_func = escape_func) for e in obj]
 	return obj
 
-def _filter_steps(steps):
-	return [(step[:1] + (_filter(step[1]), ) + step[2:]) for step in steps]
+def _escape_steps(steps, escape_func):
+	return [(step[:1] + (escape_func(step[1], escape_func = escape_func), ) + step[2:]) for step in steps]
 
-def make_pmml_pipeline(obj, active_fields = None, target_fields = None):
-	"""Translates a regular Scikit-Learn estimator or pipeline to a PMML pipeline.
+def make_pmml_pipeline(estimator, active_fields = None, target_fields = None, escape_func = _escape):
+	"""Wraps a Scikit-Learn estimator or pipeline object into a PMML pipeline object.
 
 	Parameters:
 	----------
-	obj: BaseEstimator
-		The object.
+	estimator: BaseEstimator
+		The estimator or pipeline object.
 
 	active_fields: list of strings, optional
 		Feature names. If missing, "x1", "x2", .., "xn" are assumed.
@@ -137,8 +148,12 @@ def make_pmml_pipeline(obj, active_fields = None, target_fields = None):
 		Label name(s). If missing, "y" is assumed.
 
 	"""
-	steps = _filter_steps(_get_steps(obj))
-	pipeline = PMMLPipeline(steps)
+	if not isinstance(estimator, BaseEstimator):
+		raise TypeError("The estimator object is not an instance of {0}".format(BaseEstimator.__name__))
+	pipeline = PMMLPipeline([
+		("estimator", estimator)
+	])
+	pipeline = _escape(pipeline, escape_func = escape_func)
 	if active_fields is not None:
 		pipeline.active_fields = numpy.asarray(active_fields)
 	if target_fields is not None:
@@ -175,7 +190,7 @@ def _java_version(java_home = None):
 	return _parse_java_version(error)
 
 def _parse_java_version(java_version):
-	match = re.match("^(.*)\sversion\s\"(.*)\"(|\s.+)$", java_version, re.MULTILINE)
+	match = re.match(r"^(.*)\sversion\s\"(.*)\"(|\s.+)$", java_version, re.MULTILINE)
 	if match:
 		return (match.group(1), match.group(2))
 	else:
@@ -195,7 +210,16 @@ def _process_classpath(name, fun, user_classpath):
 			else:
 				fun(zipfile.open(zipentry))
 
-def _dump(obj, prefix):
+def _dill_dump(obj, prefix):
+	fd, path = tempfile.mkstemp(prefix = (prefix + "-"), suffix = ".pkl")
+	try:
+		with open(path, "wb") as dill_file:
+			dill.dump(obj, dill_file)
+	finally:
+		os.close(fd)
+	return path
+
+def _joblib_dump(obj, prefix):
 	fd, path = tempfile.mkstemp(prefix = (prefix + "-"), suffix = ".pkl.z")
 	try:
 		joblib.dump(obj, path, compress = 3)
@@ -203,19 +227,32 @@ def _dump(obj, prefix):
 		os.close(fd)
 	return path
 
-def sklearn2pmml(pipeline, pmml, with_repr = False, java_home = None, java_opts = None, user_classpath = [], debug = False):
-	"""Converts a fitted PMML pipeline object to PMML file.
+def _is_supported(estimator):
+	clazz = estimator.__class__
+	# LightGBM and XGBoost Booster objects
+	if clazz.__module__.split(".")[0] in ["lightgbm", "xgboost"] and clazz.__name__ == "Booster":
+		return True
+	# H2O Estimator objects
+	if hasattr(estimator, "download_mojo"):
+		return True
+	return isinstance(estimator, BaseEstimator)
+
+def sklearn2pmml(estimator, pmml_path, with_repr = False, pmml_schema = None, java_home = None, java_opts = None, user_classpath = [], dump_flavour = "joblib", debug = False):
+	"""Converts a fitted estimator or pipeline object to PMML.
 
 	Parameters:
 	----------
-	pipeline: PMMLPipeline
-		The input PMML pipeline object.
+	estimator: BaseEstimator
+		The estimator or pipeline object.
 
-	pmml: string
-		The path to output PMML file.
+	pmml_path: string
+		The file path where the PMML document should be saved.
 
 	with_repr: boolean, optional
-		If true, insert the string representation of pipeline into the PMML document.
+		If true, insert a string containing a printable representation of the estimator object into the PMML document.
+
+	pmml_schema: string
+		The PMML schema version for the PMML document.
 
 	java_home: string, optional
 		The path to Java installation directory.
@@ -229,6 +266,9 @@ def sklearn2pmml(pipeline, pmml, with_repr = False, java_home = None, java_opts
 		The paths to JAR files that provide custom Transformer, Selector and/or Estimator converter classes.
 		The SkLearn2PMML classpath is constructed by appending user JAR files to package JAR files.
 
+	dump_flavour: string, optional
+		The flavour of pickle dump files.
+
 	debug: boolean, optional
 		If true, print information about the conversion process.
 
@@ -238,32 +278,46 @@ def sklearn2pmml(pipeline, pmml, with_repr = False, java_home = None, java_opts
 		if java_version is None:
 			java_version = ("java", "N/A")
 		print("python: {0}".format(platform.python_version()))
-		print("sklearn: {0}".format(sklearn.__version__))
 		print("sklearn2pmml: {0}".format(__version__))
-		print("joblib: {0}".format(joblib.__version__))
-		print("sklearn_pandas: {0}".format(sklearn_pandas.__version__))
+		print("sklearn: {0}".format(sklearn.__version__))
 		print("pandas: {0}".format(pandas.__version__))
 		print("numpy: {0}".format(numpy.__version__))
+		print("dill: {0}".format(dill.__version__))
+		print("joblib: {0}".format(joblib.__version__))
 		print("{0}: {1}".format(java_version[0], java_version[1]))
-	if not isinstance(pipeline, PMMLPipeline):
-		raise TypeError("The pipeline object is not an instance of {0}. Use the 'sklearn2pmml.make_pmml_pipeline(obj)' utility function to translate a regular Scikit-Learn pipeline or estimator to a PMML pipeline".format(PMMLPipeline.__name__))
-	if with_repr:
-		pipeline.repr_ = repr(pipeline)
+	if not _is_supported(estimator):
+		raise TypeError("The estimator object is not an instance of {0}".format(BaseEstimator.__name__))
+	# if isinstance(estimator, Pipeline):
+	if hasattr(estimator, "_final_estimator"):
+		final_estimator = estimator._final_estimator
+	else:
+		final_estimator = estimator
 	dumps = []
 	try:
 		java_args = ["-cp", os.pathsep.join(_classpath(user_classpath)), "com.sklearn2pmml.Main"]
-		estimator = pipeline._final_estimator
-		# if isinstance(estimator, H2OEstimator):
-		if hasattr(estimator, "download_mojo"):
+		if with_repr:
+			estimator.repr_ = repr(estimator)
+		# if isinstance(final_estimator, H2OEstimator):
+		if hasattr(final_estimator, "download_mojo"):
+			if dump_flavour != "dill":
+				warnings.warn("Changing dump flavour to dill")
+				dump_flavour = "dill"
 			# Avoid MOJO (re-)download if the indicator attribute is set
-			if not hasattr(estimator, "_mojo_path"):
-				estimator_mojo = estimator.download_mojo()
-				dumps.append(estimator_mojo)
-				estimator._mojo_path = estimator_mojo
-		pipeline_pkl = _dump(pipeline, "pipeline")
-		java_args.extend(["--pkl-pipeline-input", pipeline_pkl])
-		dumps.append(pipeline_pkl)
-		java_args.extend(["--pmml-output", pmml])
+			if not (hasattr(final_estimator, "_mojo_path") or hasattr(final_estimator, "_mojo_bytes")):
+				mojo_path = final_estimator.download_mojo()
+				dumps.append(mojo_path)
+				final_estimator._mojo_path = mojo_path
+		if dump_flavour == "dill":
+			pkl_path = _dill_dump(estimator, "estimator")
+		elif dump_flavour == "joblib":
+			pkl_path = _joblib_dump(estimator, "estimator")
+		else:
+			raise ValueError("Dump flavour {0} not in {1}".format(dump_flavour, ["dill", "joblib"]))
+		java_args.extend(["--pkl-input", pkl_path])
+		dumps.append(pkl_path)
+		java_args.extend(["--pmml-output", pmml_path])
+		if pmml_schema:
+			java_args.extend(["--pmml-schema", pmml_schema])
 		cmd = _make_java_command(java_home = java_home, java_opts = java_opts, java_args = java_args)
 		if debug:
 			print("Executing command:\n{0}".format(" ".join(cmd)))
@@ -274,11 +328,11 @@ def sklearn2pmml(pipeline, pmml, with_repr = False, java_home = None, java_opts
 		output, error = process.communicate()
 		retcode = process.poll()
 		if debug or retcode:
-			if len(output) > 0:
+			if len(output):
 				print("Standard output:\n{0}".format(output))
 			else:
 				print("Standard output is empty")
-			if len(error) > 0:
+			if len(error):
 				print("Standard error:\n{0}".format(error))
 			else:
 				print("Standard error is empty")
@@ -286,13 +340,13 @@ def sklearn2pmml(pipeline, pmml, with_repr = False, java_home = None, java_opts
 			raise RuntimeError("The SkLearn2PMML application has failed. The Java executable should have printed more information about the failure into its standard output and/or standard error streams")
 	finally:
 		if debug:
-			print("Preserved joblib dump file(s): {0}".format(" ".join(dumps)))
+			print("Preserved dump file(s): {0}".format(" ".join(dumps)))
 		else:
 			for dump in dumps:
 				os.remove(dump)
 
 def _parse_properties(lines):
-	splitter = re.compile("\s*=\s*")
+	splitter = re.compile(r"\s*=\s*")
 	properties = dict()
 	for line in lines:
 		line = line.decode("UTF-8").rstrip()
@@ -302,11 +356,69 @@ def _parse_properties(lines):
 		properties[key] = value
 	return properties
 
-def _supported_classes(user_classpath):
-	classes = []
-	parser = lambda x: classes.extend(_parse_properties(x.readlines()).keys())
-	_process_classpath("META-INF/sklearn2pmml.properties", parser, user_classpath)
-	return classes
+def _format_properties(properties):
+	return ["{0} = {1}\n".format(k, v) for k, v in properties.items()]
+
+def _expand_complex_key(key):
+	begin = key.find("(")
+	end = key.find(")", begin + 1)
+	if begin < 0 or end < 0:
+		return [key]
+
+	prefix = key[:begin]
+	body = key[begin + 1:end]
+	suffix = key[end + 1:]
+
+	result = []
+	parts = body.split("|")
+	for part in parts:
+		result += _expand_complex_key(prefix + part + suffix)
+	return result
+
+def _expand_mapping(mapping):
+	result = dict()
+	for k, v in mapping.items():
+		pythonClazzes = _expand_complex_key(k)
+		javaClazz = v
+		for pythonClazz in pythonClazzes:
+			result[pythonClazz] = (javaClazz if javaClazz else pythonClazz)
+	return result
+
+def load_class_mapping(user_classpath = []):
+	"""Loads the class mapping.
+
+	Parameters:
+	----------
+	user_classpath: list of strings, optional
+		The paths to JAR files that provide custom Transformer, Selector and/or Estimator converter classes.
+
+	Returns:
+	-------
+	mapping: dict
+		Mapping from Python class names to Java converter class names.
+
+	"""
+	mapping = dict()
+	func = lambda x: mapping.update(_expand_mapping(_parse_properties(x.readlines())))
+	_process_classpath("META-INF/sklearn2pmml.properties", func, user_classpath)
+	return mapping
+
+def make_class_mapping_jar(mapping, path):
+	"""Generates a class mapping JAR file.
+
+	Parameters:
+	----------
+	mapping: dict of strings
+		Mapping from Python class names to Java converter class names.
+
+	path: string
+		The path to output JAR file.
+
+	"""
+	lines = _format_properties(mapping)
+
+	with ZipFile(path, mode = "w") as zipfile:
+		zipfile.writestr("META-INF/sklearn2pmml.properties", "".join(lines))
 
 def _strip_module(name):
 	parts = name.split(".")
@@ -314,21 +426,3 @@ def _strip_module(name):
 		parts.pop(-2)
 		return ".".join(parts)
 	return name
-
-def make_tpot_pmml_config(config, user_classpath = []):
-	"""Translates a regular TPOT configuration to a PMML-compatible TPOT configuration.
-
-	Parameters:
-	----------
-	obj: config
-		The configuration dictionary.
-
-	user_classpath: list of strings, optional
-		The paths to JAR files that provide custom Transformer, Selector and/or Estimator converter classes.
-		The SkLearn2PMML classpath is constructed by appending user JAR files to package JAR files.
-
-	"""
-	tpot_keys = set(config.keys())
-	classes = _supported_classes(user_classpath)
-	pmml_keys = (set(classes)).union(set([_strip_module(class_) for class_ in classes]))
-	return { key : config[key] for key in (tpot_keys).intersection(pmml_keys)}
