@@ -1,18 +1,13 @@
 from sklearn.base import clone, BaseEstimator, ClassifierMixin, RegressorMixin
-try:
-	from sklearn.linear_model import LinearRegression
-	from sklearn.linear_model._base import LinearClassifierMixin, LinearModel, SparseCoefMixin
-except ImportError:
-	from sklearn.linear_model.base import LinearClassifierMixin, LinearModel, LinearRegression, SparseCoefMixin
+from sklearn.linear_model import LinearRegression
+from sklearn.linear_model._base import LinearClassifierMixin, LinearModel, SparseCoefMixin
 from sklearn.preprocessing import OneHotEncoder
 from sklearn.utils.metaestimators import _BaseComposition
-from sklearn2pmml.util import eval_rows
+from sklearn2pmml.util import check_predicate, eval_rows, fqn, to_expr_func
 
+import copy
 import numpy
 
-def _class_name(x):
-	return str(x.__class__)
-
 def _checkGBDTRegressor(gbdt):
 	if hasattr(gbdt, "apply"):
 		return gbdt
@@ -23,12 +18,12 @@ def _checkGBDTRegressor(gbdt):
 				return gbdt
 		except ImportError:
 			pass
-	raise ValueError("GBDT class {0} is not supported".format(_class_name(gbdt)))
+	raise TypeError("GBDT object (class {0}) is not supported".format(fqn(gbdt)))
 
 def _checkLM(lm):
 	if isinstance(lm, (LinearModel, LinearRegression, SparseCoefMixin)):
 		return lm
-	raise ValueError("LM class {0} is not supported".format(_class_name(lm)))
+	raise TypeError("LM object (class {0}) is not supported".format(fqn(lm)))
 
 def _checkGBDTClassifier(gbdt):
 	if hasattr(gbdt, "apply"):
@@ -40,14 +35,14 @@ def _checkGBDTClassifier(gbdt):
 				return gbdt
 		except ImportError:
 			pass
-	raise ValueError("GBDT class {0} is not supported".format(_class_name(gbdt)))
+	raise TypeError("GBDT object (class {0}) is not supported".format(fqn(gbdt)))
 
 def _checkLR(lr):
 	if isinstance(lr, LinearClassifierMixin):
 		return lr
-	raise ValueError("LR class {0} is not supported".format(_class_name(lr)))
+	raise TypeError("LR object (class {0}) is not supported".format(fqn(lr)))
 
-def _step_params(name, params):
+def _extract_step_params(name, params):
 	prefix = name + "__"
 	step_params = dict()
 	for k, v in (params.copy()).items():
@@ -56,6 +51,16 @@ def _step_params(name, params):
 			del params[k]
 	return step_params
 
+def _mask_params(params, mask):
+	masked_params = dict()
+	for k, v in params.items():
+		if isinstance(v, numpy.ndarray):
+			if v.shape[0] != mask.shape[0]:
+				raise ValueError()
+			v = v[mask]
+		masked_params[k] = v
+	return masked_params
+
 class GBDTEstimator(BaseEstimator):
 
 	def __init__(self, gbdt):
@@ -86,13 +91,13 @@ class GBDTLMRegressor(GBDTEstimator, RegressorMixin):
 
 	def fit(self, X, y, **fit_params):
 		self.gbdt_ = clone(self.gbdt)
-		self.gbdt_.fit(X, y, **_step_params("gbdt", fit_params))
+		self.gbdt_.fit(X, y, **_extract_step_params("gbdt", fit_params))
 		id = self._leaf_indices(X)
 		self.ohe_ = OneHotEncoder(categories = "auto")
 		self.ohe_.fit(id)
 		idt = self.ohe_.transform(id)
 		self.lm_ = clone(self.lm)
-		self.lm_.fit(idt, y, **_step_params("lm", fit_params))
+		self.lm_.fit(idt, y, **_extract_step_params("lm", fit_params))
 		return self
 
 	def predict(self, X):
@@ -105,15 +110,19 @@ class GBDTLRClassifier(GBDTEstimator, ClassifierMixin):
 		super(GBDTLRClassifier, self).__init__(_checkGBDTClassifier(gbdt))
 		self.lr = _checkLR(lr)
 
+	@property
+	def classes_(self):
+		return self.lr_.classes_
+
 	def fit(self, X, y, **fit_params):
 		self.gbdt_ = clone(self.gbdt)
-		self.gbdt_.fit(X, y, **_step_params("gbdt", fit_params))
+		self.gbdt_.fit(X, y, **_extract_step_params("gbdt", fit_params))
 		id = self._leaf_indices(X)
 		self.ohe_ = OneHotEncoder(categories = "auto")
 		self.ohe_.fit(id)
 		idt = self.ohe_.transform(id)
 		self.lr_ = clone(self.lr)
-		self.lr_.fit(idt, y, **_step_params("lr", fit_params))
+		self.lr_.fit(idt, y, **_extract_step_params("lr", fit_params))
 		return self
 
 	def predict(self, X):
@@ -124,15 +133,34 @@ class GBDTLRClassifier(GBDTEstimator, ClassifierMixin):
 		idt = self._encoded_leaf_indices(X)
 		return self.lr_.predict_proba(idt)
 
+def _to_sparse(X, step_mask, step_result):
+	# Make array
+	if len(step_result.shape) == 1:
+		result = numpy.empty((X.shape[0], ), dtype = object)
+	else:
+		result = numpy.empty((X.shape[0], step_result.shape[1]), dtype = object)
+	# Fill array
+	if len(step_result.shape) == 1:
+		result[step_mask.ravel()] = step_result
+	else:
+		result[step_mask.ravel(), :] = step_result
+	return result
+
 class _BaseEnsemble(_BaseComposition):
 
-	def __init__(self, steps):
+	def __init__(self, steps, controller):
 		for step in steps:
 			if type(step) is not tuple:
-				raise ValueError("Step is not a tuple")
+				raise TypeError("Step is not a tuple")
 			if len(step) != 3:
-				raise ValueError("Step is not a three-element (name, estimator, predicate) tuple")
+				raise TypeError("Step is not a three-element (name, estimator, predicate) tuple")
+			name, estimator, predicate = step
+			check_predicate(predicate)
 		self.steps = steps
+		if controller:
+			if not hasattr(controller, "transform"):
+				raise TypeError()
+		self.controller = controller
 
 	@property
 	def _steps(self):
@@ -149,18 +177,10 @@ class _BaseEnsemble(_BaseComposition):
 		self._set_params("_steps", **kwargs)
 		return self
 
-def _to_sparse(X, step_mask, step_result):
-	# Make array
-	if len(step_result.shape) == 1:
-		result = numpy.empty((X.shape[0], ), dtype = object)
-	else:
-		result = numpy.empty((X.shape[0], step_result.shape[1]), dtype = object)
-	# Fill array
-	if len(step_result.shape) == 1:
-		result[step_mask.ravel()] = step_result
-	else:
-		result[step_mask.ravel(), :] = step_result
-	return result
+	def _to_evaluation_dataset(self, X):
+		if self.controller is not None:
+			return self.controller.transform(X)
+		return X
 
 class Link(BaseEstimator):
 
@@ -169,16 +189,19 @@ class Link(BaseEstimator):
 		self.augment_funcs = augment_funcs
 		self.prefit = prefit
 		if prefit:
-			self.estimator_ = clone(self.estimator)
+			self.estimator_ = copy.deepcopy(self.estimator)
 
-	def fit(self, X, y):
+	def fit(self, X, y, **fit_params):
 		self.estimator_ = clone(self.estimator)
-		self.estimator_.fit(X, y)
+		self.estimator_.fit(X, y, **fit_params)
 		return self
 
 	def predict(self, X):
 		return self.estimator_.predict(X)
 
+	def predict_proba(self, X):
+		return self.estimator_.predict_proba(X)
+
 	def augment(self, X):
 		Y = None
 		for augment_func in self.augment_funcs:
@@ -195,8 +218,8 @@ class Link(BaseEstimator):
 
 class EstimatorChain(_BaseEnsemble):
 
-	def __init__(self, steps, multioutput = True):
-		super(EstimatorChain, self).__init__(steps)
+	def __init__(self, steps, controller = None, multioutput = True):
+		super(EstimatorChain, self).__init__(steps, controller)
 		self.multioutput = multioutput
 
 	def fit(self, X, y, **fit_params):
@@ -204,30 +227,37 @@ class EstimatorChain(_BaseEnsemble):
 			if len(self.steps) != y.shape[1]:
 				raise ValueError()
 			y = numpy.asarray(y)
+		X_eval = self._to_evaluation_dataset(X)
 		i = 0
 		for name, estimator, predicate in self.steps:
-			step_mask = eval_rows(X, lambda X: eval(predicate), dtype = bool)
+			step_mask_func = to_expr_func(predicate)
+			step_mask = eval_rows(X_eval, step_mask_func, dtype = bool)
 			if numpy.sum(step_mask) < 1:
 				raise ValueError(predicate)
+			step_X = X[step_mask]
 			if len(y.shape) == 1:
 				step_y = y[step_mask]
 			elif len(y.shape) == 2:
 				step_y = y[step_mask, i]
 			else:
 				raise ValueError()
-			estimator.fit(X[step_mask], step_y, **_step_params(name, fit_params))
+			step_fit_params = _mask_params(_extract_step_params(name, fit_params), step_mask)
+			estimator.fit(step_X, step_y, **step_fit_params)
 			if isinstance(estimator, Link):
 				X = estimator.augment(X)
 			i += 1
 		return self
 
-	def predict(self, X):
+	def _predict(self, X, predict_method):
 		result = None
+		X_eval = self._to_evaluation_dataset(X)
 		for name, estimator, predicate in self.steps:
-			step_mask = eval_rows(X, lambda X: eval(predicate), dtype = bool)
+			step_mask_func = to_expr_func(predicate)
+			step_mask = eval_rows(X_eval, step_mask_func, dtype = bool)
 			if numpy.sum(step_mask) < 1:
 				continue
-			step_result = estimator.predict(X[step_mask])
+			step_X = X[step_mask]
+			step_result = getattr(estimator, predict_method)(step_X)
 			step_result = _to_sparse(X, step_mask, step_result)
 			if self.multioutput:
 				step_result = step_result.reshape(X.shape[0], -1)
@@ -237,36 +267,58 @@ class EstimatorChain(_BaseEnsemble):
 				if self.multioutput:
 					result = numpy.hstack((result, step_result))
 				else:
+					if step_result.shape != result.shape:
+						raise ValueError()
 					result[step_mask] = step_result[step_mask]
 			if isinstance(estimator, Link):
 				X = estimator.augment(X)
 		return result
 
+	def predict(self, X):
+		return self._predict(X, "predict")
+
+	def predict_proba(self, X):
+		return self._predict(X, "predict_proba")
+
 class SelectFirstEstimator(_BaseEnsemble):
 
-	def __init__(self, steps):
-		super(SelectFirstEstimator, self).__init__(steps)
+	def __init__(self, steps, controller, eval_rows):
+		super(SelectFirstEstimator, self).__init__(steps, controller)
+		self.eval_rows = eval_rows
+
+	def _eval_step_mask(self, X, predicate):
+		step_func = to_expr_func(predicate)
+		if self.eval_rows:
+			return eval_rows(X, step_func, dtype = bool)
+		else:
+			return step_func(X)
 
 	def fit(self, X, y, **fit_params):
-		mask = numpy.zeros(X.shape[0], dtype = bool)
+		X_eval = self._to_evaluation_dataset(X)
+		mask = numpy.full(X.shape[0], fill_value = False)
 		for name, estimator, predicate in self.steps:
-			step_mask = eval_rows(X, lambda X: eval(predicate), dtype = bool)
+			step_mask = self._eval_step_mask(X_eval, predicate)
 			step_mask[mask] = False
 			if numpy.sum(step_mask) < 1:
 				raise ValueError(predicate)
-			estimator.fit(X[step_mask], y[step_mask], **_step_params(name, fit_params))
+			step_X = X[step_mask]
+			step_y = y[step_mask]
+			step_fit_params = _mask_params(_extract_step_params(name, fit_params), step_mask)
+			estimator.fit(step_X, step_y, **step_fit_params)
 			mask = numpy.logical_or(mask, step_mask)
 		return self
 
-	def _predict(self, predict_method, X):
+	def _predict(self, X, predict_method):
 		result = None
-		mask = numpy.zeros(X.shape[0], dtype = bool)
+		X_eval = self._to_evaluation_dataset(X)
+		mask = numpy.full(X.shape[0], fill_value = False)
 		for name, estimator, predicate in self.steps:
-			step_mask = eval_rows(X, lambda X: eval(predicate), dtype = bool)
+			step_mask = self._eval_step_mask(X_eval, predicate)
 			step_mask[mask] = False
 			if numpy.sum(step_mask) < 1:
 				continue
-			step_result = getattr(estimator, predict_method)(X[step_mask])
+			step_X = X[step_mask]
+			step_result = getattr(estimator, predict_method)(step_X)
 			step_result = _to_sparse(X, step_mask, step_result)
 			if result is None:
 				result = step_result
@@ -276,20 +328,20 @@ class SelectFirstEstimator(_BaseEnsemble):
 		return result
 
 	def apply(self, X):
-		return self._predict("apply", X)
+		return self._predict(X, "apply")
 
 	def predict(self, X):
-		return self._predict("predict", X)
+		return self._predict(X, "predict")
 
 class SelectFirstRegressor(SelectFirstEstimator, RegressorMixin):
 
-	def __init__(self, steps):
-		super(SelectFirstRegressor, self).__init__(steps)
+	def __init__(self, steps, controller = None, eval_rows = True):
+		super(SelectFirstRegressor, self).__init__(steps, controller, eval_rows)
 
 class SelectFirstClassifier(SelectFirstEstimator, ClassifierMixin):
 
-	def __init__(self, steps):
-		super(SelectFirstClassifier, self).__init__(steps)
+	def __init__(self, steps, controller = None, eval_rows = True):
+		super(SelectFirstClassifier, self).__init__(steps, controller, eval_rows)
 
 	def predict_proba(self, X):
-		return self._predict("predict_proba", X)
+		return self._predict(X, "predict_proba")
